from .conf import connection, construct_redis_key
from .exceptions import TaskDoesNotExist
from .registries import queue_registry
from .task import Task
from .utils import atomic_pipeline, decode_list


class Queue(object):
    def __init__(self, name='default'):
        self.name = name
        self.key = construct_redis_key('queue:' + name)
        # We use a separate key for the workers to wait on, as we need to do a
        # multi-key blocking rpop on it, and redis does not have a variant of
        # that operation that is not at risk of losing tasks.
        self.unblock_key = construct_redis_key('unblock_queue:' + name)

    @classmethod
    def all(cls):
        """Returns an iterable of all Queues."""
        return [cls(name) for name in sorted(queue_registry.get_names())]

    def count(self):
        return connection.llen(self.key)

    def _empty_transaction(self, pipeline):
        task_ids = decode_list(pipeline.lrange(self.key, 0, -1))
        pipeline.multi()
        Task.delete_many(task_ids, pipeline=pipeline)
        pipeline.delete(self.key)
        pipeline.delete(self.unblock_key)

    def empty(self):
        """Removes all messages on the queue."""
        connection.transaction(self._empty_transaction, self.key)

    def delete(self):
        def transaction(pipeline):
            self._empty_transaction(pipeline)
            queue_registry.remove(self, pipeline=pipeline)
        connection.transaction(transaction, self.key)

    def get_task_ids(self, offset=0, length=-1):
        end = -1 - offset
        start = -(length + 1) if length < 0 else end - length
        return [task_id.decode() for task_id in
                reversed(connection.lrange(self.key, start, end))]

    def get_tasks(self, offset=0, length=-1):
        return Task.fetch_many(self.get_task_ids(offset, length))

    @atomic_pipeline
    def enqueue_call(self, *args, pipeline, **kwargs):
        """Creates a task to represent the delayed function call and enqueues it."""
        task = Task(*args, **kwargs)
        task.enqueue(self, pipeline=pipeline)
        return task

    @atomic_pipeline
    def push(self, task, *, pipeline, at_front=False):
        """Pushes a task on the queue

        `at_front` inserts the task at the front instead of the back of the queue"""
        queue_registry.add(self, pipeline=pipeline)
        if at_front:
            pipeline.rpush(self.key, task.id)
        else:
            pipeline.lpush(self.key, task.id)
        pipeline.lpush(self.unblock_key, task.id)

    def remove_and_delete(self, task):
        def transaction(pipeline):
            task_ids = decode_list(pipeline.lrange(self.key, 0, -1))
            if task.id not in task_ids:
                raise TaskDoesNotExist()

            pipeline.multi()
            pipeline.lrem(self.key, 0, task.id)
            Task.delete_many([task.id], pipeline=pipeline)

        connection.transaction(transaction, self.key)

    def dequeue(self, worker):
        """Dequeue a task and set it as the current task for `worker`"""
        # Use lua script to atomically clear unblock_key if queue is empty
        lua = connection.register_script("""
            local queue, unblocker, worker_task_list = unpack(KEYS)
            local result = redis.call("RPOPLPUSH", queue, worker_task_list)
            if result == false then
                redis.call("DEL", unblocker)
            end
            return result
        """)
        result = lua(keys=[self.key, self.unblock_key, worker.task_key])
        if result is None:
            return None
        else:
            task_id = result.decode()
            worker.current_task_id = task_id
            return Task.fetch(task_id)

    @classmethod
    def await_multi(cls, queues, timeout):
        """Blocks until one of the passed queues contains a tasks.

        Return the queue that contained a task or None if `timeout` was reached."""
        queue_map = {q.unblock_key: q for q in queues}
        result = connection.brpop(queue_map.keys(), timeout)
        if result is None:
            return None
        return queue_map[result[0].decode()]

    def __eq__(self, other):
        if type(other) != type(self):
            return NotImplemented
        else:
            return self.name == other.name

    def __hash__(self):
        return hash((self.__class__.__name__, self.name))

    def __repr__(self):
        return '{0}({1!r})'.format(self.__class__.__name__, self.name)

    def __str__(self):
        return '<{0} {1}>'.format(self.__class__.__name__, self.name)



import pytest

from redis_tasks import exceptions, utils


def test_import_attribute():
    assert utils.import_attribute("tests.app.import_me.test_attrib") == "foo"


def test_utc():
    now = utils.utcnow()
    assert utils.utcparse(utils.utcformat(now)) == now.replace(microsecond=0)


def test_serialization():
    my_obj = {'a': 'b', 'c': 54.325, 'd': utils.utcnow()}
    assert utils.deserialize(utils.serialize(my_obj)) == my_obj

    with pytest.raises(exceptions.DeserializationError):
        utils.deserialize(b"invalid")


def test_atomic_pipeline(mocker):
    connection = mocker.patch('redis_tasks.conf.connection')

    def f1(pipeline):
        assert pipeline == mock_pipe

    mock_pipe = mocker.sentinel.pipe
    utils.atomic_pipeline(f1)(pipeline=mock_pipe)

    def f2(pipeline):
        assert pipeline == real_pipe

    real_pipe = connection.pipeline().__enter__()
    utils.atomic_pipeline(f2)()
    real_pipe.execute.assert_called_once()

    def f3(pipeline):
        raise ZeroDivisionError()

    real_pipe.execute.reset_mock()
    with pytest.raises(ZeroDivisionError):
        utils.atomic_pipeline(f3)()
    real_pipe.execute.assert_not_called()



import multiprocessing
import os
import signal

import pytest

from redis_tasks.exceptions import WorkerShutdown
from redis_tasks.task import Task
from redis_tasks.worker_process import PostponeShutdown, WorkHorse
from tests.utils import TaskFactory


def test_postpone_shutdown():
    PostponeShutdown._shutdown_delayed = False
    with pytest.raises(WorkerShutdown):
        PostponeShutdown.trigger_shutdown()

    ps = PostponeShutdown()
    ps.__enter__()
    PostponeShutdown.trigger_shutdown()
    with pytest.raises(WorkerShutdown):
        ps.__exit__()

    PostponeShutdown._shutdown_delayed = False
    ps = PostponeShutdown()
    ps.activate()
    PostponeShutdown.trigger_shutdown()
    with pytest.raises(WorkerShutdown):
        ps.deactivate()

    PostponeShutdown._shutdown_delayed = False
    ps1 = PostponeShutdown()
    ps1.activate()
    with PostponeShutdown():
        PostponeShutdown.trigger_shutdown()
    with pytest.raises(WorkerShutdown):
        ps1.deactivate()

    PostponeShutdown._shutdown_delayed = False
    assert not PostponeShutdown._active


def test_run(mocker):
    task = TaskFactory()
    execute = mocker.patch.object(task, 'execute', return_value=mocker.sentinel.outcome)
    conn1, conn2 = multiprocessing.Pipe()
    WorkHorse(task, conn2).run()
    execute.assert_called_once()
    assert conn1.poll()
    assert conn1.recv() is True
    assert conn1.poll()
    assert conn1.recv().name == 'outcome'
    assert conn1.poll() is False

    # do not leave the PostponeShutdown active
    with execute.call_args[1]["shutdown_cm"]:
        pass


def horsetask():
    raise ArithmeticError("work me")


def test_process():
    task = Task(horsetask)
    horse_conn, conn2 = multiprocessing.Pipe()
    horse = WorkHorse(task, conn2)
    horse.start()
    horse.join()
    assert horse_conn.poll()
    assert horse_conn.recv() is True
    assert horse_conn.poll()
    outcome = horse_conn.recv()
    assert horse_conn.poll() is False
    assert outcome.outcome == 'failure'
    assert outcome.message.splitlines()[-1] == 'ArithmeticError: work me'


def horsetaskwait():
    with multiprocessing.connection.Client(os.environ['RT_TEST_SOCKET']) as conn:
        conn.send("A")
        assert conn.recv() == "B"
        conn.send("C")
        conn.recv()


@pytest.fixture()
def suprocess_socket(tmpdir):
    socket_file = str(tmpdir.join('socket'))
    os.environ['RT_TEST_SOCKET'] = socket_file
    with multiprocessing.connection.Listener(socket_file) as listener:
        yield listener


def test_process_shutdown(suprocess_socket):
    # sanity checks against leftovers from previous tests
    assert not PostponeShutdown._active
    assert not PostponeShutdown._shutdown_delayed

    task = Task(horsetaskwait)
    horse_conn, conn2 = multiprocessing.Pipe()
    horse = WorkHorse(task, conn2)
    horse.start()
    assert horse_conn.poll(1)
    assert horse_conn.recv() is True
    with suprocess_socket.accept() as taskconn:
        assert taskconn.poll(1)
        assert taskconn.recv() == "A"
        # horse ignores SIGTERM
        horse.send_signal(signal.SIGTERM)
        taskconn.send("B")
        assert taskconn.poll(1)
        assert taskconn.recv() == "C"
        # Trigger shutdown
        horse.send_signal(signal.SIGUSR1)
        horse.join(1)
        assert not horse.is_alive()
    assert horse_conn.poll()
    outcome = horse_conn.recv()
    assert not horse_conn.poll()
    assert outcome.outcome == 'failure'
    assert 'Worker shutdown' in outcome.message.splitlines()[-1]



import datetime
import random
import uuid

import pytest

from redis_tasks.exceptions import InvalidOperation, TaskDoesNotExist
from redis_tasks.registries import (
    failed_task_registry, finished_task_registry, worker_registry)
from redis_tasks.task import Task, TaskOutcome, TaskStatus, redis_task
from redis_tasks.utils import decode_list
from tests.utils import QueueFactory, WorkerFactory


class SomeClass:
    def some_func(self):
        pass

    def misleading_func(self):
        pass


def misleading_func():
    pass


def name_func():
    pass


@redis_task(reentrant=True)
def reentrant_stub():
    pass


def test_init(stub):
    def closure():
        pass

    with pytest.raises(ValueError):
        Task(closure)
    with pytest.raises(ValueError):
        Task(SomeClass.some_func)
    with pytest.raises(ValueError):
        Task(SomeClass().some_func)
    with pytest.raises(ValueError):
        Task(SomeClass().misleading_func)

    with pytest.raises(TypeError):
        Task(stub, args="foo")
    with pytest.raises(TypeError):
        Task(stub, kwargs=["foo"])

    t = Task(name_func, ["foo"], {"bar": "a"})
    assert t.description == "tests.test_task.name_func('foo', bar='a')"


def test_state_transistions(assert_atomic, connection, time_mocker, stub):
    time = time_mocker('redis_tasks.task.utcnow')
    task = Task(reentrant_stub)
    q = QueueFactory()
    w = WorkerFactory()
    w.startup()

    # enqueue
    time.step()
    assert not connection.exists(task.key)
    with assert_atomic():
        task.enqueue(q)
    assert q.get_task_ids() == [task.id]
    assert connection.exists(task.key)
    for t in [task, Task.fetch(task.id)]:
        assert t.enqueued_at == time.now
        assert t.status == TaskStatus.QUEUED
        assert t.origin == q.name

    # dequeue
    task = q.dequeue(w)
    assert q.get_task_ids() == []
    assert worker_registry.get_running_tasks() == {w.id: task.id}

    # set_running
    time.step()
    with assert_atomic():
        w.start_task(task)
    assert worker_registry.get_running_tasks() == {w.id: task.id}
    for t in [task, Task.fetch(task.id)]:
        assert t.status == TaskStatus.RUNNING
        assert t.started_at == time.now

    # requeue
    time.step()
    with assert_atomic():
        w.end_task(task, TaskOutcome("requeue"))
    assert worker_registry.get_running_tasks() == dict()
    assert q.get_task_ids() == [task.id]
    for t in [task, Task.fetch(task.id)]:
        assert t.status == TaskStatus.QUEUED
        assert t.started_at is None

    # set_finished
    task = q.dequeue(w)
    w.start_task(task)
    time.step()
    with assert_atomic():
        w.end_task(task, TaskOutcome("success"))
    assert q.get_task_ids() == []
    assert finished_task_registry.get_task_ids() == [task.id]
    for t in [task, Task.fetch(task.id)]:
        assert t.status == TaskStatus.FINISHED
        assert t.ended_at == time.now

    # set_failed
    task = q.enqueue_call(stub)
    task = q.dequeue(w)
    w.start_task(task)
    assert worker_registry.get_running_tasks() == {w.id: task.id}
    time.step()
    with assert_atomic():
        w.end_task(task, TaskOutcome("failure", message="my error"))
    assert worker_registry.get_running_tasks() == dict()
    assert failed_task_registry.get_task_ids() == [task.id]
    for t in [task, Task.fetch(task.id)]:
        assert t.status == TaskStatus.FAILED
        assert t.error_message == "my error"
        assert t.ended_at == time.now


def test_cancel(assert_atomic, connection):
    q = QueueFactory()
    task = q.enqueue_call()
    with assert_atomic():
        task.cancel()
    assert q.get_task_ids() == []
    assert task.status == TaskStatus.CANCELED
    assert not connection.exists(task.key)

    w = WorkerFactory()
    w.startup()
    task = q.enqueue_call()
    q.dequeue(w)
    with pytest.raises(InvalidOperation):
        with assert_atomic():
            task.cancel()
    assert worker_registry.get_running_tasks() == {w.id: task.id}
    assert connection.exists(task.key)


def test_worker_death(assert_atomic, connection, stub):
    def setup(func):
        w = WorkerFactory()
        w.startup()
        q.enqueue_call(func)
        task = q.dequeue(w)
        return task, w

    q = QueueFactory()

    # Worker died before starting work on the task
    task, w = setup(stub)
    assert worker_registry.get_running_tasks() == {w.id: task.id}
    with assert_atomic(exceptions=['hgetall']):
        w.died()
    assert q.get_task_ids() == [task.id]
    assert worker_registry.get_running_tasks() == dict()
    assert failed_task_registry.get_task_ids() == []

    # Worker died after starting non-reentrant task
    q.empty()
    task, w = setup(stub)
    assert worker_registry.get_running_tasks() == {w.id: task.id}
    w.start_task(task)
    with assert_atomic(exceptions=['hgetall']):
        w.died()
    assert q.get_task_ids() == []
    assert worker_registry.get_running_tasks() == dict()
    assert failed_task_registry.get_task_ids() == [task.id]

    # Worker died after starting reentrant task
    connection.delete(failed_task_registry.key)
    task, w = setup(reentrant_stub)
    assert worker_registry.get_running_tasks() == {w.id: task.id}
    with assert_atomic():
        w.start_task(task)
    with assert_atomic(exceptions=['hgetall']):
        w.died()
    assert q.get_task_ids() == [task.id]
    assert worker_registry.get_running_tasks() == dict()
    assert failed_task_registry.get_task_ids() == []


def test_get_func(stub):
    assert Task(stub)._get_func() == stub


@redis_task(timeout=42)
def my_timeout_func():
    pass


def test_get_properties(settings, stub):
    assert not Task(stub).is_reentrant
    assert Task(reentrant_stub).is_reentrant

    assert Task(stub).timeout == settings.DEFAULT_TASK_TIMEOUT
    assert Task(my_timeout_func).timeout == 42


def test_delete_many(connection, assert_atomic):
    tasks = [Task() for i in range(5)]
    for t in tasks:
        t._save()

    with assert_atomic():
        Task.delete_many([tasks[0].id, tasks[1].id])

    for t in tasks[0:2]:
        assert not connection.exists(t.key)
    for t in tasks[2:5]:
        assert connection.exists(t.key)


def test_persistence(assert_atomic, connection, stub):
    fields = {'func_name', 'args', 'kwargs', 'status', 'origin', 'description',
              'error_message', 'enqueued_at', 'started_at', 'ended_at', 'meta',
              'aborted_runs'}

    def randomize_data(task):
        string_fields = ['func_name', 'status', 'description', 'origin', 'error_message']
        date_fields = ['enqueued_at', 'started_at', 'ended_at']
        for f in string_fields:
            setattr(task, f, str(uuid.uuid4()))
        for f in date_fields:
            setattr(task, f, datetime.datetime(
                random.randint(1000, 9999), 1, 1, tzinfo=datetime.timezone.utc))

        task.args = tuple(str(uuid.uuid4()) for i in range(4))
        task.kwargs = {str(uuid.uuid4()): ["d"]}
        task.meta = {"x": [str(uuid.uuid4())]}
        task.aborted_runs = ["foo", "bar", str(uuid.uuid4())]

    def as_dict(task):
        return {f: getattr(task, f) for f in fields}

    task = Task(stub)
    with assert_atomic():
        task._save()
    assert set(decode_list(connection.hkeys(task.key))) <= fields
    assert as_dict(Task.fetch(task.id)) == as_dict(task)

    randomize_data(task)
    task._save()
    assert as_dict(Task.fetch(task.id)) == as_dict(task)

    # only deletes
    task.enqueued_at = None
    task.error_message = None
    task._save(['enqueued_at', 'error_message'])
    assert task.enqueued_at is None
    assert task.error_message is None

    for i in range(5):
        store = random.sample(fields, 7)
        copy = Task.fetch(task.id)
        randomize_data(copy)
        copy._save(store)
        for f in store:
            setattr(task, f, getattr(copy, f))
        assert as_dict(Task.fetch(task.id)) == as_dict(task)

    copy = Task.fetch(task.id)
    randomize_data(copy)
    copy.meta = task.meta = {"new_meta": "here"}
    copy.save_meta()
    copy.refresh()
    assert as_dict(copy) == as_dict(task)
    assert as_dict(Task.fetch(task.id)) == as_dict(task)

    # save() and save_meta() in same pipeline
    with assert_atomic():
        with connection.pipeline() as pipe:
            task = Task(stub)
            task._save(pipeline=pipe)
            task.meta["a"] = "b"
            task.save_meta(pipeline=pipe)
            pipe.execute()
    assert Task.fetch(task.id).meta == {"a": "b"}

    task = Task(stub)
    with pytest.raises(TaskDoesNotExist):
        task.refresh()
    with pytest.raises(TaskDoesNotExist):
        Task.fetch('nonexist')


def test_init_save_fetch_delete(connection, assert_atomic, stub):
    t = Task(stub, ["foo"])
    assert not connection.exists(t.key)
    with assert_atomic():
        t._save()
    assert connection.exists(t.key)

    fetched = Task.fetch(t.id)
    assert fetched.id == t.id
    assert fetched.args == ["foo"]

    Task.delete_many([t.id])
    assert not connection.exists(t.key)
    with pytest.raises(TaskDoesNotExist):
        Task.fetch(t.id)


def test_fetch_many(stub):
    tasks_data = [{"args": []}, {"args": ["foo"]}, {"args": ["bar"]},
                  {"args": ["foo", "bar"]}, {"args": []}]
    tasks = [Task(stub, **d) for d in tasks_data]
    for t in tasks:
        t._save()

    fetched = Task.fetch_many([t.id for t in tasks[1:4]])
    assert len(fetched) == 3
    for i, task in enumerate(fetched):
        assert task.id == tasks[i+1].id
        assert task.args == tasks_data[i+1]["args"]



import pytest

from redis_tasks import registries
from redis_tasks.conf import construct_redis_key
from redis_tasks.exceptions import WorkerDoesNotExist
from redis_tasks.task import TaskOutcome
from tests.utils import QueueFactory, Something, TaskFactory, WorkerFactory


def test_expiring_registry(connection, settings, mocker, assert_atomic):
    registry = registries.ExpiringRegistry('testexpire')
    task1 = TaskFactory()
    task2 = TaskFactory()
    settings.EXPIRING_REGISTRIES_TTL = 10
    delete_tasks = mocker.patch('redis_tasks.task.Task.delete_many')

    assert registry.key == construct_redis_key('testexpire_tasks')

    timestamp = mocker.patch('redis_tasks.conf.RTRedis.time')
    timestamp.return_value = (1000, 0)
    with assert_atomic():
        registry.add(task1)
    timestamp.return_value = (1004, 0)
    registry.add(task2)

    registry.expire()
    assert registry.get_task_ids() == [task1.id, task2.id]

    timestamp.return_value = (1012, 0)
    registry.expire()
    assert registry.get_task_ids() == [task2.id]
    delete_tasks.assert_called_once_with([task1.id], pipeline=Something)


def test_worker_registry(connection, settings, mocker, assert_atomic):
    registry = registries.worker_registry
    worker1 = WorkerFactory()
    worker2 = WorkerFactory()
    settings.WORKER_HEARTBEAT_TIMEOUT = 100

    timestamp = mocker.patch('redis_tasks.conf.RTRedis.time')
    timestamp.return_value = (1000, 0)
    with assert_atomic():
        registry.add(worker1)
    timestamp.return_value = (1050, 0)
    registry.add(worker2)
    assert registry.get_worker_ids() == [worker1.id, worker2.id]

    timestamp.return_value = (1120, 0)
    assert registry.get_dead_ids() == [worker1.id]

    registry.heartbeat(worker1)
    timestamp.return_value = (1140, 0)
    assert registry.get_dead_ids() == []

    timestamp.return_value = (1200, 0)
    assert registry.get_dead_ids() == [worker2.id]

    with assert_atomic():
        registry.remove(worker1)
    # worker is already dead
    with pytest.raises(WorkerDoesNotExist):
        registry.heartbeat(worker1)

    with assert_atomic():
        with connection.pipeline() as pipe:
            registry.remove(worker1, pipeline=pipe)
            registry.remove(worker2, pipeline=pipe)
            pipe.execute()
    assert registry.get_worker_ids() == []
    assert registry.get_dead_ids() == []


def test_worker_reg_running_tasks():
    registry = registries.worker_registry
    queue = QueueFactory()
    t1 = queue.enqueue_call()
    t2 = queue.enqueue_call()
    worker1 = WorkerFactory(queues=[queue])
    worker2 = WorkerFactory(queues=[queue])
    worker1.startup()
    worker2.startup()

    assert registry.get_running_tasks() == dict()
    queue.push(t1)
    queue.dequeue(worker1)
    assert registry.get_running_tasks() == {worker1.id: t1.id}
    worker1.start_task(t1)

    queue.push(t2)
    queue.dequeue(worker2)
    worker2.start_task(t2)
    assert registry.get_running_tasks() == {worker1.id: t1.id, worker2.id: t2.id}
    worker1.end_task(t1, TaskOutcome("success"))
    assert registry.get_running_tasks() == {worker2.id: t2.id}


def test_queue_registry(assert_atomic, connection):
    registry = registries.queue_registry
    queue1 = QueueFactory()
    queue2 = QueueFactory()

    with assert_atomic():
        registry.add(queue1)
    registry.add(queue2)
    assert set(registry.get_names()) == {queue1.name, queue2.name}

    with assert_atomic():
        registry.remove(queue1)
    assert registry.get_names() == [queue2.name]
    registry.remove(queue2)
    assert registry.get_names() == []


def test_maintenance():
    registries.registry_maintenance()



import multiprocessing
import os
import signal
import time

import pytest

from redis_tasks.exceptions import WorkerShutdown
from redis_tasks.task import TaskOutcome, TaskStatus
from redis_tasks.utils import decode_list
from redis_tasks.worker import WorkerState
from redis_tasks.worker_process import (
    ShutdownRequested, WorkerProcess, WorkHorse, generate_worker_description)
from tests.utils import QueueFactory, TaskFactory, id_list


def test_generate_description():
    assert generate_worker_description()


def dummy_name():
    return "dummy"


def test_init(assert_atomic, settings):
    settings.WORKER_DESCRIPTION_FUNCTION = "tests.test_workerprocess.dummy_name"

    wp = WorkerProcess([QueueFactory])
    assert wp.worker.description == "dummy"


def test_queue_iter(connection, mocker):
    queues = [QueueFactory(), QueueFactory()]
    wp = WorkerProcess(queues)
    wp.worker.startup()
    qi = wp.queue_iter(False)
    task1 = queues[1].enqueue_call()
    task2 = queues[0].enqueue_call()
    assert next(qi).id == task2.id
    assert next(qi).id == task1.id

    def my_await_multi(*args):
        nonlocal await_counter, task3, task4, task5
        await_counter += 1
        if await_counter == 1:
            return None
        elif await_counter == 2:
            task3 = queues[1].enqueue_call()
            return queues[1]
        elif await_counter == 3:
            task4 = queues[1].enqueue_call()
            task5 = queues[1].enqueue_call()
            return queues[0]
        assert False

    await_counter = 0
    task3 = task4 = task5 = None
    mocker.patch('redis_tasks.queue.Queue.await_multi', new=my_await_multi)
    assert next(qi).id == task3.id
    assert wp.worker.current_task_id == task3.id
    assert await_counter == 2
    assert next(qi).id == task4.id
    assert wp.worker.current_task_id == task4.id
    assert await_counter == 3
    assert next(qi).id == task5.id
    assert await_counter == 3

    # Assert that tasks are moved to the worker
    assert (decode_list(connection.lrange(wp.worker.task_key, 0, -1)) ==
            id_list([task5, task4, task3, task1, task2]))

    # Burst mode
    wp = WorkerProcess(queues)
    wp.worker.startup()
    task1 = queues[1].enqueue_call()
    task2 = queues[0].enqueue_call()
    task3 = queues[1].enqueue_call()
    assert id_list(wp.queue_iter(True)) == id_list([task2, task1, task3])


def test_run(settings, mocker, stub):
    tasks = [mocker.sentinel.t1, mocker.sentinel.t2, mocker.sentinel.t3]
    mocker.patch('redis_tasks.worker_process.WorkerProcess.queue_iter', return_value=tasks)

    def my_process(task):
        assert wp.worker.state == WorkerState.IDLE

    process = mocker.patch('redis_tasks.worker_process.WorkerProcess.process_task',
                           side_effect=my_process)
    maintenance = mocker.patch('redis_tasks.worker_process.Maintenance')
    wp = WorkerProcess([QueueFactory()])
    assert wp.run(True) == 3
    assert process.call_args_list == [((x, ),) for x in tasks]
    assert maintenance().run_if_neccessary.call_count == 4
    assert wp.worker.state == WorkerState.DEAD

    settings.WORKER_PRELOAD_FUNCTION = stub.path
    stub.mock.reset_mock()
    wp = WorkerProcess([QueueFactory()])
    wp.run(True)
    assert stub.mock.call_count == 1


def test_run_shutdown(settings, mocker):
    mocker.patch.object(WorkerProcess, 'queue_iter', side_effect=ShutdownRequested)
    wp = WorkerProcess([QueueFactory()])
    with pytest.raises(ShutdownRequested):
        wp.run(False)
    assert wp.worker.started_at
    assert wp.worker.shutdown_at


def test_heartbeats(mocker, stub):
    heartbeat_sent = False

    def send_heartbeat(*args, **kwargs):
        nonlocal heartbeat_sent
        heartbeat_sent = True
    mocker.patch('redis_tasks.registries.WorkerRegistry.heartbeat', side_effect=send_heartbeat)

    def consume_heartbeat(*args, **kwargs):
        nonlocal heartbeat_sent
        assert heartbeat_sent is True
        heartbeat_sent = False
        return mocker.DEFAULT
    maintenance = mocker.patch('redis_tasks.worker_process.Maintenance.run_if_neccessary',
                               side_effect=consume_heartbeat)

    def my_await(*args):
        nonlocal task
        for i in range(3):
            consume_heartbeat()
            print("Awaited!")
            yield None
        task = queue.enqueue_call(stub)
        consume_heartbeat()
        yield queue
        raise ShutdownRequested()
    mock_await = mocker.patch('redis_tasks.queue.Queue.await_multi', side_effect=my_await())

    mocker.patch.object(WorkHorse, 'start', new=WorkHorse.run)
    horse_alive = mocker.patch.object(WorkHorse, 'is_alive', return_value=True)

    def my_join(*args):
        consume_heartbeat()
        yield None
        consume_heartbeat()
        yield None
        consume_heartbeat()
        horse_alive.return_value = False
        yield None
    mock_join = mocker.patch.object(WorkHorse, 'join', side_effect=my_join())

    stub.mock.reset_mock()
    queue = QueueFactory()
    task = None
    wp = WorkerProcess([queue])
    with pytest.raises(ShutdownRequested):
        wp.run(False)
    task.refresh()
    assert task.status == TaskStatus.FINISHED
    assert maintenance.call_count == 2
    assert mock_await.call_count == 5
    assert mock_join.call_count == 3
    assert stub.mock.called


def test_process_task(mocker):
    q = QueueFactory()
    wp = WorkerProcess([q])
    wp.worker.startup()
    q.enqueue_call()
    task = q.dequeue(wp.worker)

    def my_execute(task):
        assert wp.worker.state == WorkerState.BUSY
        assert task.status == TaskStatus.RUNNING
        return TaskOutcome("success")

    execute = mocker.patch.object(WorkerProcess, 'execute_task', side_effect=my_execute)
    wp.process_task(task)
    assert task.status == TaskStatus.FINISHED

    q.enqueue_call()
    task = q.dequeue(wp.worker)
    execute.side_effect = ArithmeticError()
    wp.process_task(task)
    assert task.status == TaskStatus.FAILED
    assert 'ArithmeticError' in task.error_message


def test_execute_task(mocker, settings, time_mocker):
    horse = None

    def my_start(self):
        nonlocal horse
        horse = self
        assert horse.task == task
        horse_alive.return_value = True
        horse.worker_connection.send(True)

    mocker.patch.object(WorkHorse, 'start', new=my_start)
    horse_alive = mocker.patch.object(WorkHorse, 'is_alive', return_value=False)

    # Normal run
    def my_join():
        yield None
        yield None
        horse.worker_connection.send(TaskOutcome("success"))
        horse_alive.return_value = False
        yield None

    mock_join = mocker.patch.object(WorkHorse, 'join', side_effect=my_join())
    wp = WorkerProcess([QueueFactory()])
    wp.worker.startup()
    task = TaskFactory()
    outcome = wp.execute_task(task)
    assert outcome.outcome == "success"
    assert mock_join.call_count == 3

    # Unexpected WorkHorse death
    def dying_join():
        yield None
        horse_alive.return_value = False
        yield None

    mock_join.side_effect = dying_join()
    outcome = wp.execute_task(task)
    assert outcome.outcome == "failure"
    assert "Workhorse died unexpectedly" in outcome.message

    # Shutdown
    shutdown_initiated = False

    def take_usr1(signum):
        nonlocal shutdown_initiated
        shutdown_initiated = True
        assert signum == signal.SIGUSR1

    def shutdown_join():
        yield None
        assert wp.in_interruptible
        yield ShutdownRequested()
        assert shutdown_initiated
        yield None
        horse.worker_connection.send(TaskOutcome("requeue"))
        yield None
        horse_alive.return_value = False
        yield None
    mock_join.side_effect = shutdown_join()
    fake_signal = mocker.patch.object(WorkHorse, 'send_signal', side_effect=take_usr1)
    outcome = wp.execute_task(task)
    assert outcome.outcome == "requeue"

    # Timeout
    def timeout_join():
        yield None
        yield None
        time.step()
        yield None
        yield None

    def take_kill(signum):
        assert signum == signal.SIGKILL
        horse_alive.return_value = False

    mock_join.side_effect = timeout_join()
    settings.DEFAULT_TASK_TIMEOUT = 1
    time = time_mocker("redis_tasks.worker_process.utcnow")
    time.step()
    fake_signal = mocker.patch.object(WorkHorse, 'send_signal', side_effect=take_kill)
    outcome = wp.execute_task(task)
    assert outcome.outcome == "failure"
    assert "Task timeout (1 sec) reached" in outcome.message
    fake_signal.assert_called_once_with(9)


def taskwait():
    with multiprocessing.connection.Client(os.environ['RT_TEST_SOCKET']) as conn:
        conn.send("A")
        try:
            time.sleep(10)
            conn.send("C")
        except WorkerShutdown:
            conn.send("B")
            raise


@pytest.fixture()
def suprocess_socket(tmpdir):
    socket_file = str(tmpdir.join('socket'))
    os.environ['RT_TEST_SOCKET'] = socket_file
    with multiprocessing.connection.Listener(socket_file) as listener:
        yield listener


def test_signal_shutdown_in_task(suprocess_socket):
    queue = QueueFactory()
    task = queue.enqueue_call(taskwait)
    wp = WorkerProcess([queue])
    process = multiprocessing.Process(target=wp.run)
    process.start()

    with suprocess_socket.accept() as taskconn:
        assert taskconn.poll(1)
        assert taskconn.recv() == "A"
        os.kill(process.pid, signal.SIGTERM)
        assert taskconn.poll(1)
        assert taskconn.recv() == "B"
        process.join(1)
        assert not process.is_alive()
    task.refresh()
    assert task.status == TaskStatus.FAILED
    assert 'Worker shutdown' in task.error_message.splitlines()[-1]


def test_signal_shutdown_in_queuewait():
    wp = WorkerProcess([QueueFactory()])
    process = multiprocessing.Process(target=wp.run)
    process.start()
    time.sleep(0.1)
    os.kill(process.pid, signal.SIGTERM)
    process.join(1)
    assert not process.is_alive()



from contextlib import contextmanager

from raven.transport.threaded import ThreadedHTTPTransport

from ..conf import settings
from ..utils import LazyObject, import_attribute

sentry = LazyObject(lambda: import_attribute(settings.SENTRY_INSTANCE))


def set_client(client):
    global sentry
    sentry = client


class SentryMiddleware:
    @contextmanager
    def context(self, task):
        sentry.context.activate()
        sentry.context.merge({'extra': {
            'task_id': task.id,
            'task_func': task.func_name,
            'task_args': task.args,
            'task_kwargs': task.kwargs,
            'task_description': task.description,
        }})
        sentry.transaction.push(task.func_name)
        try:
            yield
        finally:
            sentry.transaction.pop(task.func_name)
            sentry.context.clear()

    def run_task(self, task, run, args, kwargs):
        with self.context(task):
            run(*args, **kwargs)

    def process_outcome(self, task, *exc_info):
        try:
            if not exc_info[0]:
                return
            with self.context(task):
                sentry.captureException(exc_info=exc_info)
        finally:
            self.wait_for_messages()

    def wait_for_messages(self):
        transport = sentry.remote.get_transport()
        if isinstance(transport, ThreadedHTTPTransport):
            worker = transport.get_worker()
            worker._timed_queue_join(10)



class TaskAborted(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class TaskDoesNotExist(Exception):
    pass


class InvalidOperation(Exception):
    pass


class WorkerDoesNotExist(Exception):
    pass


class DeserializationError(Exception):
    def __init__(self, message, raw_data):
        super().__init__(message)
        self.raw_data = raw_data


class WorkerShutdown(BaseException):
    pass
